/*
HikiDoc.js - JavaScript implementation of 'HikiDoc' which is a
text-to-HTML conversion tool.

The original 'HikiDoc' was written by Kazuhiko <kazuhiko@fdiary.net>.

This release was made by KAWABATA, Kazumichi (higemaru)
<kawabata@cpan.org> http://haro.jp/

Copyright (c) 2011- KAWABATA, Kazumichi (higemaru)


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.
    * Neither the name of the HikiDoc nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
var HikiDoc = function(args) {
    var self = this;
    self.brMode = true;
    self.hLevel = 1;

    if ( args && args.brMode == false )
        self.brMode = false;
    if ( args && args.hLevel ) {
        var tmp = parseInt(args.hLevel);
        if ( tmp > 1 && tmp <= 6 )
            self.hLevel = parseInt(args.hLevel);
    }

    self.toHtml = function(str) {
        if ( typeof(str) !== 'string' )
            return '';
        else if ( str === '' )
            return '';

        var _hikidocSTACK = new Array();
        var _pluginSTACK = new Array();

        // get line feed code, and unite
        var brCode = getBrCode(str);
        str = str.replace(/\r\n|\r|\n/g, "\n");
        str = str.replace(/\n*$/,"\n\n");
        // escape html
        str = self.escapeHTML(str);
        // escape some meta char
        str = self._escapeMetaChar(str);

        // parse blocks
        str = _blockParser(str);

        // remove needless new lines
        str = str.replace(/\n{2,}/g, "\n");

        // restore some html parts
        str = _restoreBlock(str);
        str = _restorePluginBlock(str);

        // unescape some meta char
        str = _unescapeMetaChar(str);

        // terminate with a single new line
        str = str.replace(/\n*$/g, '');

	str = str.replace(/\n/g, brCode);

        return str;

        function getBrCode(s) {
            if ( s.indexOf("\r\n") > 0 )
                return "\r\n";
	    else if ( s.indexOf("\r") > 0 )
                return "\r";
	    else
                return "\n";
        }
        function chomp(s) {
            return s.replace(new RegExp(/\n$/), '');
        }
        function trim(s) {
            return s.replace(new RegExp(/\n$/), '').replace(new RegExp(/^\n/), '');
        }
        function _unescapeMetaChar(s,f) {
            if ( s == undefined )
                s = '';
            if ( f && f == true )
                s = s.replace(/&#(\d+);/g,
                              function(dum,ch){return '\\' + String.fromCharCode(ch);}
                              );
            else
                s = s.replace(/&#(\d+);/g,
                              function(dum,ch){return String.fromCharCode(ch);}
                              );

            return s;
        }

        function _storeBlock(s) {
            if ( s == undefined )
                s = '';
            _hikidocSTACK.push(s);
            return '<' + parseInt(_hikidocSTACK.length - 1) + '>';
        }
        function _restoreBlock(s,c) {
            if ( s == undefined )
                s = '';
            if ( c == undefined )
                c = 0;

            // recursive < 10
            if ( _hikidocSTACK.length <= 0 || c > 10)
                return s;

            s = s.replace(/<(\d+)>/gm, function(dum,ch){;return _restoreBlock(_hikidocSTACK[ch], c+1);} );

            return s;
        }
        function _storePluginBlock(s) {
            if ( s == undefined )
                s = '';
            _pluginSTACK.push(s);
            return '&lt;!' + parseInt(_pluginSTACK.length - 1) + '&gt;';
        }
        function _restorePluginBlock(s,f) {
            if ( s == undefined )
                s = '';
            if ( _pluginSTACK.length < 0 )
                return s;

            if ( f && f == true )
                s = s.replace(/&lt;!(\d+)&gt;/g, function(matchedStr, num){
                        return _pluginSTACK[num];
                    });
            else {
                s = s.replace(/<p>&lt;!(\d+)&gt;<\/p>/g, function(matchedStr, num){
                        return '<div class="plugin">' + _pluginSTACK[num] + '</div>';
                    });
                s = s.replace(/&lt;!(\d+)&gt;/g, function(matchedStr, num){
                        return '<span class="plugin">' + _pluginSTACK[num] + '</span>';
                    });
            }

            return s;
        }

        //
        // inline parser
        //
        function _inlineParser(s) {

            // parse link and image
            s = _parseLink(s);
            s = _parseModifier(s);

            return s;

            function _parseLink(s) {
                // blacket [[ ]] -> link
                s = s.replace(/\[\[(.+?)\]\]/gm,
                              function(matchedStr,str){
                                  var uri = ''; var title = '';
                                  if ( str.match(/(.+?)\|(.+)/) ) {
                                      title = RegExp.$1;
                                      uri = RegExp.$2;
                                      title = _parseModifier(title);
                                  }
                                  else {
                                      title = str;
                                      uri = str;
                                  }
                                  if ( ! uri.match(/\:\/\//)
                                       && ! uri.match(/^mailto:/) )
                                      uri = uri.replace(/^(?:https?|ftp|file)+:/, '');
                                  var key = _storeBlock('<a href="' + uri.replace(/\"/g,'&quot;') + '">' + title + '</a>');
                                  return matchedStr.replace(/\[\[(.+?)\]\]/, key);
                              }
                              );

                // naked string -> link
                s = s.replace(/((?:(?:https?|ftp|file):|mailto:)[A-Za-z0-9;\/?:@&=+$,\-_.!~*\'()#%]+)/gm,
                              function(str){
                                  var uri = str;
                                  if ( ! uri.match(/\:\/\//)
                                       && ! uri.match(/^mailto:/) )
                                      uri = uri.replace(/^\w+:/, '');

                                  var key = '';
                                  if ( uri.match(/\.(?:jpe?g|gif|png)$/) ) { // images
                                      var basename = uri.replace(/^.*\/([^/]+)$/, function(dum,c){ return c; });
                                      key = _storeBlock('<img src="' + uri + '" alt="' + basename + '" />');
                                  }
                                  else
                                      key = _storeBlock('<a href="' + uri + '">' + uri + '</a>');

                                  return str.replace(/((?:(?:https?|ftp|file):|mailto:)[A-Za-z0-9;\/?:@&=+$,\-_.!~*\'()#%]+)/m, key);

                              }
                              );

                return s;
            }

            function _parseModifier(s) {
                if ( s == undefined )
                    s = '';

                var _strong = '\'\'\''; var _em = '\'\''; var _del = '==';
                var _strongRegStr = _strong + '(.+?)' + _strong;
                var _emRegStr = _em + '(.+?)' + _em;
                var _delRegStr = _del + '(.+?)' + _del;
                var modRegStr = '(' + _strongRegStr + '|' + _emRegStr + '|' + _delRegStr + ')';
                modRegStr = modRegStr.replace(/\(\.\+\?\)/, '(?:.+?)');
                var modReg    = new RegExp(modRegStr, 'gm');
                s = s.replace(modReg,
                              function(matchedStr){
                                  var ret = matchedStr;
                                  ret = ret.replace(new RegExp('(.*)' + _strongRegStr + '(.*)'), function(m,bstr,str,astr) {
                                          return _storeBlock(_parseModifier(bstr + '<strong>' + _parseModifier(str) + '</strong>' + astr));
                                      });
                                  ret = ret.replace(new RegExp('(.*)' + _emRegStr + '(.*)'), function(m,bstr,str,astr) {
                                          return _storeBlock(_parseModifier(bstr + '<em>' + _parseModifier(str) + '</em>' + astr));
                                      });
                                  ret = ret.replace(new RegExp('(.*)' + _delRegStr + '(.*)'), function(m,bstr,str,astr) {
                                          return _storeBlock(_parseModifier(bstr + '<del>' + _parseModifier(str) + '</del>' + astr));
                                      });
                                  return ret;
                              }
                    );

                return s;
            }

        }


        function _blockParser(s) {
            // parse plugin
            var strTmp = '';
            var pStrTmp = '';
            var isPlugin = false;

	    //var pTmps = s.split(/({{|}})/); // mmm.... ie split...
	    var pTmps = new Array;
	    var pTmps1 = s.split('{{');
	    for ( var i = 0; i < pTmps1.length; i++ ) {
		var pTmps2 = pTmps1[i].split('}}');
		for ( var j = 0; j < pTmps2.length; j++ ) {
		    pTmps.push(pTmps2[j]);
		    if ( pTmps2[j+1] )
			pTmps.push('}}');
		}
		if ( pTmps1[i+1] )
		    pTmps.push('{{');
	    }

            for ( var i = 0; i < pTmps.length; i++ ) {
                var str = pTmps[i];
                if ( str == '{{' ) {
                    isPlugin = true;
                    pStrTmp = pStrTmp + str;
                }
                else if ( str == '}}') {
                    if ( isPlugin ) {
                        pStrTmp = pStrTmp + str;
                        var rTmp = pStrTmp.replace(/(\'|\")[\s\S]*?\1/g, '');
                        if ( ! rTmp.match(/(?:\'|\")/) ) {
                            isPlugin = false;
                            strTmp = strTmp + _storePluginBlock( _unescapeMetaChar(pStrTmp, true) );
                            pStrTmp = '';
                        }
                    }
                    else {
                        strTmp = strTmp + str;
                    }
                }
                else {
                    if ( isPlugin )
                        pStrTmp = pStrTmp + str;
                    else
                        strTmp = strTmp + str;
                }
            }
            if ( isPlugin )
                strTmp = strTmp + pStrTmp;
            s = strTmp;

            // parse pre
            // <<< - >>>
            s = s.replace(/^&lt;&lt;&lt;[ \t]*(\w*)$([\s\S]*?)^\&gt;\&gt;\&gt;$/gm,
                          function(matchedStr,preOpt,preInner) {
                              return "\n" + _storeBlock('<pre>' + _restorePre(preInner) + '</pre>') + "\n\n";
                          }
                          );
            // space at the head
            s = s.replace(/(?:^[ \t].*\n?)+/gm,
                          function(matchedStr) {
                              var preInner = chomp(matchedStr);
                              preInner = preInner.replace(/^[ \t]/gm,'');
                              return "\n" + _storeBlock("<pre>\n" + _restorePre(preInner) + "\n</pre>") + "\n\n";
                          }
                          );
            function _restorePre(s) {
                s = _unescapeMetaChar(s, true);
                s = _restorePluginBlock(s, true);
                return s;
            }

            // parse comment
            s = s.replace(/^\/\/.*\n?/gm, '');
            // parse header
            var re = new RegExp('^(!{1,'+parseInt(7-self.hLevel)+'}) *(.*)\n?','gm');
            s = s.replace(re,
                          function(matchedStr,level,str){
                              var i = parseInt(level.length + self.hLevel -1);
                              str = _inlineParser(str);

                              return "\n<h" + i + '>' + str + '</h' + i + ">\n\n"; }
                          );
            // parse hrules
            s = s.replace(/^----$/gm, '\n<hr />\n');

            // parse list
            var _ul = '*';
            var _ol = '#';
            var _listMark = '[' + _ul + _ol + ']';
            var listsReg = new RegExp('(?:^' + _listMark + '+ *.*\n)+', 'gm');
            var listReg = new RegExp('^((' + _listMark + ')+) *(.*)', 'gm');
            s = s.replace( listsReg,
                           function(str) {
                               var listTypeArray = new Array();
                               var curStr = "\n";
                               var level = 0;
                               str = str.replace( listReg,
                                                  function(str,newL,listT,listItem) {
                                                      var newLevel = newL.length;
                                                      var listType = (listT == _ul) ? 'ul' : 'ol';

                                                      if ( newLevel > level ) {
                                                          for ( var i = 0; i < newLevel - level; i++ ) {
                                                              listTypeArray.push(listType);
                                                              curStr = curStr + '<' + listType + ">\n<li>";
                                                          }
                                                          curStr = curStr + _inlineParser(listItem);
                                                      }
                                                      else if ( newLevel < level ) {
                                                          for ( var i = 0; i < level - newLevel; i++ ) {
                                                              curStr = curStr + "</li>\n</" + listTypeArray.pop() + '>';
                                                          }
                                                          curStr = curStr + "</li>\n<li>" + _inlineParser(listItem);
                                                      }
                                                      else if ( listType == listTypeArray[listTypeArray.length - 1] ) {
                                                          curStr = curStr + "</li>\n<li>" + _inlineParser(listItem);
                                                      }
                                                      else {
                                                          curStr = curStr + "</li>\n</" + listTypeArray.pop() + ">\n";
                                                          curStr = curStr + '<' + listType + ">\n";
                                                          curStr = curStr + '<li>' + _inlineParser(listItem);
                                                          listTypeArray.push(listType);
                                                      }
                                                      level = newLevel;
                                                      return str;
                                                  }
                                                  );
                               for ( var i = 0; i < level; i++ ) {
                                   curStr = curStr + "</li>\n</" + listTypeArray.pop() + '>';
                               }
                               curStr = curStr + "\n\n";

                               return curStr;
                           }
                           );

            // parse definition
            var _definition = '^:(.*?)?:(.*)\n?';
            var _definitions = '(?:' + _definition + ')+';

            s = s.replace(new RegExp(_definition,'gm'), function(c){return _inlineParser(c);});

            s = s.replace(new RegExp(_definitions, 'gm'), function(str) {
                    str = chomp(str);
                    str = str.replace(new RegExp(_definition, 'gm'), function(st,dt,dd) {
                            var ret = '';
                            if ( dt == '' || dt == undefined )
                                ret = ret + '<dd>' + dd + "</dd>\n";
                            else if ( dd == '' || dd == undefined )
                                ret = ret + '<dt>' + dt + "</dt>\n";
                            else
                                ret = ret +  '<dt>' + dt + '</dt><dd>' + dd + "</dd>\n";
                            return ret;
                        });
                    return "\n<dl>\n" + str + "<\/dl>\n\n";
                });

            // parse blockquote
            var _bq  = '^\"\"[ \t]?';
            var bqsReg = new RegExp('((?:' + _bq + '.*\n?)+)','gm');
            s = s.replace(bqsReg,
                          function(str){
                              str = chomp(str);
                              str = str.replace(new RegExp(_bq,'gm'), '');
                              return "\n<blockquote>\n" + _blockParser(str) + "\n</blockquote>\n\n";
                          }
                          );
            // parse table
            var _table = '\|\|';
            var _tables = '(?:^' + _table + '.+\n?)+';

            s = s.replace(/^\|\|.+\n?/gm, function(str) { return _inlineParser(str); } );
            s = s.replace(/(?:^\|\|.+\n?)+/gm, function(str) {
                    str = str.replace(/^\|\|.+$/gm, function(str) {
                            str = chomp(str);
                            str = str.replace(/^\|\|/, '');
                            var ret = '';
                            var cols = str.split('||');
                            var colsLen = cols.length -1;
                            if ( cols[colsLen] ) colsLen ++;
                            for (var i = 0; i < colsLen; i++) {
                                var isTh = 0;
                                var col = cols[i].replace(/^\!/,function(){isTh++;return '';});
                                var th = isTh ? 'th' : 'td';
                                var attr = '';
                                col = col.replace(/^(\^|&gt;)+/, function(str,sp) {
                                        var span = ( sp == '^' ) ? 'row' : 'col';
                                        var rowLen = str.split('^').length;
                                        var colLen = str.split('&gt;').length;
                                        if ( rowLen > 1 )
                                            attr = attr + ' rowspan="' + rowLen + '"';
                                        if ( colLen > 1 )
                                            attr = attr + ' colspan="' + colLen + '"';

                                        return '';
                                    });
                                ret = ret + '<' + th + attr + '>' + _inlineParser(col) + '</' + th + '>';
                            }
                            return '<tr>' + ret + "</tr>";
                        });
                    return "<table>\n" + str + "</table>\n\n";
                });

            // parse paragraph
            var paraTmps = s.split(/\n{2,}/); // mmmm. ie6...
            var paraLen = paraTmps.length
            var retArray = new Array();
            for ( var i = 0; i < paraLen; i++ ) {
                var p = paraTmps[i];
                var str = trim(p);

                if ( str == '' )
                    retArray.push('');
                else if ( str.match(/^\<[^!]/) ) {
                    retArray.push(str);
                }else {
                    str = '<p>' + _inlineParser(str) + '</p>';
                    if ( self.brMode === true)
                        str = str.replace(/\n/g, function(){ return "<br />\n"; } );
                    retArray.push(str);
                }
            }
            s = retArray.join("\n\n");
            s = s.replace(/^ +/gm, '');

            return s;
        }
    }
    self.escapeHTML = function(s) {
        var o = document.createElement('pre');
////        if ( typeof(o.textContent) !== undefined )
//          if ( o.textContent )
//              o.textContent = s;
//          else // for IE6
//              o.innerText = s;
//          return o.innerHTML;
        return s.replace(/&/g,'&amp;').replace(/</g,'&lt;').replace(/>/g,'&gt;');
    }
    self._escapeMetaChar = function(s) {
        // \{, \}, \:, \', \", \|
        return s.replace(/\\({|}|:|\'|\"|\|)/g,
                         function(dum,ch){return '&#' + ch.charCodeAt(0) + ';'}
                         );
    }

}
